C%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
C
C GENDATA3D generate randomly distributed coordinate and charge
C data for an arbitrary number of particles. The user is prompted
C for N, the number of particles, and the choice of particle distribution:
C
C  (1) uniformly distributed data in a [-0.5,0.5]^3 CUBE
C  (2) uniformly distributed data in a [-0.5,0.5]^3 cube which is then
C      projected radially onto the surface of the unit SPHERE
C
C%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

      PROGRAM GENDATA3D
      IMPLICIT NONE      

C r8 is 8-byte (double precision) real
   
      INTEGER, PARAMETER :: r8=SELECTED_REAL_KIND(12)

C runtime parameter

      INTEGER :: numpars

C arrays for coordinates and charge

      REAL(KIND=r8), ALLOCATABLE, DIMENSION(:) :: x,y,z,q

C local variable

      INTEGER :: i,ichoice,err
      REAL(KIND=r8) :: r

C EXECUTABLE STATEMENTS

      WRITE(6,*) 'Enter NUMPARS : '
      READ(5,*) numpars
      WRITE(6,*) ' Cube data (1) or surface of a Sphere (2) :'
      READ(5,*) ichoice

      ALLOCATE(x(numpars),y(numpars),z(numpars),q(numpars),STAT=err)
      IF(err .NE. 0) THEN
         WRITE(*,*) 'Error allocating arrays for x, y, z, and q! '
         STOP
      ENDIF
      
      CALL RANDOM_NUMBER(x)
      CALL RANDOM_NUMBER(y)
      CALL RANDOM_NUMBER(z)
      CALL RANDOM_NUMBER(q)  
      
      x = x-0.5_r8
      y = y-0.5_r8
      z = z-0.5_r8
      q = q-0.5_r8

      IF (ichoice .EQ. 1) THEN
      
         WRITE(6,*) ' '
         WRITE(6,*) 'CUBE: uniformly distributed data in [-0.5,0.5]^3'
         WRITE(6,*) ' '

         OPEN(unit=80,file='xyz_cube.dat')
         DO i=1,numpars
            WRITE(80,*) x(i),y(i),z(i)
         END DO
         CLOSE(unit=80)
      
      ELSE

         WRITE(6,*) ' '
         WRITE(6,*) 'SPHERE: uniformly distributed data in [-0.5,0.5]^3'
         WRITE(6,*) 'projected radially onto surface of the unit SPHERE'
         WRITE(6,*) ' '       
         
         DO i = 1,numpars
            r = SQRT(x(i)*x(i)+y(i)*y(i)+z(i)*z(i))
            x(i) = x(i)/r
            y(i) = y(i)/r
            z(i)  = z(i)/r
         END DO

         OPEN(unit=80,file='xyz_sphere.dat')
         DO i=1,numpars
            WRITE(80,*) x(i),y(i),z(i)
         END DO
         CLOSE(unit=80)
         
      END IF  

      OPEN(unit=81,file='q.dat')
      DO i = 1,numpars
         WRITE(81,*) q(i)
      END DO
      CLOSE(unit=81)

      DEALLOCATE(x,y,z,q,STAT=err)
      IF(err .NE. 0) THEN
         WRITE(*,*) 'Error deallocating arrays for x, y, z, and q! '
         STOP
      ENDIF

      END PROGRAM GENDATA3D