      PROGRAM TREEDRIVER_YUKAWA
      IMPLICIT NONE

C r8 is 8-byte (double precision) real

      INTEGER,PARAMETER :: r8=SELECTED_REAL_KIND(12)

C runtime parameters

      INTEGER :: numpars,order,maxparnode,iflag,forcedim
      REAL(KIND=r8) :: kappa,theta 

C arrays for coordinates, charge, potential & force (tree and direct) 

      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:) :: x,y,z,q
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:) :: tpoten,dpoten
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:,:) :: tforce,dforce
      INTEGER,ALLOCATABLE,DIMENSION(:) :: orderind

C timing variables

      REAL(KIND=r8) :: timebeg,timeend

C local variables

      INTEGER :: i,j,err
      REAL(KIND=r8) :: t1,abserr,relerr,absinf_err,relinf_err
      REAL(KIND=r8),DIMENSION(3) :: f_inferr,f_relinferr,t

C BEGIN EXECUTABLE STATEMENTS

      WRITE(6,*) 'Enter NUMPARS: '
      READ(5,*) numpars
      WRITE(6,*) 'Enter KAPPA: '
      READ(5,*) kappa
      WRITE(6,*) 'Enter THETA: '
      READ(5,*) theta
      WRITE(6,*) 'Enter ORDER: '
      READ(5,*) order
      WRITE(6,*) 'Enter MAXPARNODE: '
      READ(5,*) maxparnode
      WRITE(6,*) 'Enter IFLAG  (1-potential energy '
      WRITE(6,*) '             (2-potential energy and force) : '
      READ(5,*) iflag
      
      WRITE(6,*) ' '
      WRITE(6,*) 'Run parameters: '
      WRITE(6,*) '                 numpars    = ',numpars
      WRITE(6,*) '                 kappa      = ',kappa
      WRITE(6,*) '                 theta      = ',theta
      WRITE(6,*) '                 order      = ',order
      WRITE(6,*) '                 maxparnode = ',maxparnode
      WRITE(6,*) '                 iflag      = ',iflag      

      ALLOCATE(x(numpars),y(numpars),z(numpars),q(numpars),
     &         orderind(numpars),STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error allocating x, y, z, q, or orderind!'
          STOP
      END IF
      ALLOCATE(tpoten(numpars),dpoten(numpars),STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error allocating tpoten or dpoten!'
          STOP
      END IF
      
      IF (iflag .EQ. 1) THEN
         forcedim=1
      ELSE
         forcedim=numpars
      END IF
      ALLOCATE(tforce(forcedim,3),dforce(forcedim,3),STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error allocating tforce or dforce! '
          STOP
      END IF      

      OPEN(unit=80,file='xyz.dat')
      OPEN(unit=81,file='q.dat')
      DO i=1,numpars
         READ(80,*) x(i),y(i),z(i)
         READ(81,*) q(i)
      END DO
      CLOSE(unit=80)
      CLOSE(unit=81)
      
C compute potential or potential&force using treecode

      CALL TREECODE3D_YUKAWA(x,y,z,q,kappa,order,theta,maxparnode,
     &                       numpars,orderind,iflag,tpoten,tforce,
     &                       forcedim)

C compute potential or potential&force directly
      
      IF (iflag .EQ. 1) THEN
         WRITE(6,*) ' '
         WRITE(6,*) 'Computing potential - directly'
      ELSE
         WRITE(6,*) ' '
         WRITE(6,*) 'Computing potential and force - directly'
      END IF
      
      CALL CPU_TIME(timebeg)
      CALL COMPUTE_DIRECT(x,y,z,q,kappa,numpars,dpoten,dforce,
     &                    forcedim,iflag)      
      CALL CPU_TIME(timeend)

      WRITE(6,*) 'Time for direct computation (secs):',timeend-timebeg

C compute errors

      WRITE(6,*) ' '
      WRITE(6,*) 'Computing potential error'
      WRITE(6,*) ' '
      
      abserr = 0.0_r8
      relerr = 0.0_r8
      relinf_err = 0.0_r8
      absinf_err = 0.0_r8
      DO j=1,numpars
         t1 = ABS(dpoten(j)-tpoten(j))
         relerr = relerr+t1**2
         abserr = abserr+dpoten(j)*dpoten(j)
         IF (t1 .GT. relinf_err) THEN
            relinf_err = t1
         END IF
         t1 = ABS(dpoten(j))
         IF (t1 .GT. absinf_err) THEN
            absinf_err = t1
         END IF
      END DO
      
      relerr = SQRT(relerr/abserr)
      relinf_err =  relinf_err/absinf_err

      WRITE(6,*) '   Relative L2 and Inf error :',relerr,relinf_err
      WRITE(6,*) ' '

      IF (iflag .EQ. 2) THEN

         WRITE(6,*) ' '
         WRITE(6,*) 'Computing force component errors'

C compute force errors

         f_inferr = 0.0_r8
         f_relinferr = 0.0_r8

         DO j = 1,numpars
            DO i = 1,3
               t(i) = dforce(j,i)-tforce(j,i)
               IF (ABS(t(i)) .GT. f_inferr(i)) THEN
                  f_inferr(i) = ABS(t(i))
               END IF
            END DO
         END DO
         f_relinferr(1) = f_inferr(1)/MAXVAL(ABS(dforce(1:numpars,1)))
         f_relinferr(2) = f_inferr(2)/MAXVAL(ABS(dforce(1:numpars,2)))
         f_relinferr(3) = f_inferr(3)/MAXVAL(ABS(dforce(1:numpars,3)))

         WRITE(6,*) ' '
         WRITE(6,*) '   Absolute Inf norm force component error:'      
         WRITE(6,*) ' '
         WRITE(6,*) '   ',f_inferr(1),f_inferr(2),f_inferr(3)
         WRITE(6,*) ' '
         WRITE(6,*) '   Relative Inf norm force component error:'     
         WRITE(6,*) ' '
         WRITE(6,*) '   ',f_relinferr(1),f_relinferr(2),f_relinferr(3)
         WRITE(6,*) ' '

      END IF
          
      DEALLOCATE(x,y,z,q,orderind,STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error deallocating x, y, z, q, or orderind!'
          STOP
      END IF
      DEALLOCATE(tpoten,dpoten,STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error deallocating tpoten or dpoten!'
          STOP
      END IF
      
      DEALLOCATE(tforce,dforce,STAT=err)
      IF (err .NE. 0) THEN
          WRITE(6,*) 'Error deallocating tforce or dforce! '
          STOP
      END IF          

      END PROGRAM TREEDRIVER_YUKAWA

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      SUBROUTINE COMPUTE_DIRECT(x,y,z,q,kappa,numpars,dpoten,dforce,
     &                  forcedim,iflag)
      IMPLICIT NONE

      INTEGER,PARAMETER :: r8=SELECTED_REAL_KIND(12)
      INTEGER,INTENT(IN) :: numpars,iflag,forcedim
      REAL(KIND=r8),DIMENSION(numpars),INTENT(IN) :: x,y,z,q
      REAL(KIND=r8),DIMENSION(forcedim,3),INTENT(INOUT) :: dforce
      REAL(KIND=r8),DIMENSION(numpars),INTENT(INOUT) :: dpoten
      REAL(KIND=r8),INTENT(IN) :: kappa

C local variables 
   
      INTEGER :: i,j
      REAL(KIND=r8) :: tx,ty,tz,fx,fy,fz,teng,dist,t1
      REAL(KIND=r8) ::  dpeng,temp,peng

      IF (iflag .NE. 1) THEN

         dforce(1:numpars,:) = 0.0_r8
         dpoten = 0.0_r8
         DO i = 1,numpars-1
            peng = 0.0_r8
            fx = 0.0_r8
            fy = 0.0_r8
            fz = 0.0_r8        
            DO j = i+1,numpars
               tx = x(j)-x(i)
               ty = y(j)-y(i)
               tz = z(j)-z(i)
               dist = SQRT(tx*tx+ty*ty+tz*tz)
               temp = EXP(-kappa*dist)/dist
               peng = peng + q(j)*temp
               dpoten(j) = dpoten(j) + q(i)*temp
               temp = temp*(kappa*dist+1.0_r8)/(dist*dist)
               tx = -tx*temp
               ty = -ty*temp
               tz = -tz*temp
               fx = fx + q(j)*tx
               fy = fy + q(j)*ty
               fz = fz + q(j)*tz
               dforce(j,1) = dforce(j,1) - q(i)*tx
               dforce(j,2) = dforce(j,2) - q(i)*ty
               dforce(j,3) = dforce(j,3) - q(i)*tz       
            END DO
            dpoten(i) = q(i)*(dpoten(i)+peng)
            dforce(i,1) = q(i)*(dforce(i,1)+fx)
            dforce(i,2) = q(i)*(dforce(i,2)+fy)
            dforce(i,3) = q(i)*(dforce(i,3)+fz)
         END DO
         dpoten(numpars) = q(numpars)*dpoten(numpars)         
         dforce(numpars,1) = q(numpars)*dforce(numpars,1)
         dforce(numpars,2) = q(numpars)*dforce(numpars,2)
         dforce(numpars,3) = q(numpars)*dforce(numpars,3)
         
      ELSE
      
         dforce = 0.0_r8
         dpoten = 0.0_r8
         DO i = 1,numpars-1
            peng = 0.0_r8
            DO j = i+1,numpars
               tx = x(j)-x(i)
               ty = y(j)-y(i)
               tz = z(j)-z(i)
               dist = SQRT(tx*tx+ty*ty+tz*tz)
               temp = EXP(-kappa*dist)/dist
               peng = peng + q(j)*temp
               dpoten(j) = dpoten(j) + q(i)*temp
            END DO
            dpoten(i) = q(i)*(dpoten(i)+peng)
         END DO
         dpoten(numpars) = q(numpars)*dpoten(numpars)      
      
      END IF
 
      RETURN
      END SUBROUTINE COMPUTE_DIRECT     


